from argparse import RawDescriptionHelpFormatter

from .command import OasisBaseCommand
from ..utils.inputs import InputValues

from .. import __version__


class ConfigUpdateCmd(OasisBaseCommand):
    """
    Read in an MDK config file and writes an updated file, replacing deprecated keys
    with newer ones compatible with the current MDK release
    """

    def add_args(self, parser):
        """
        Adds arguments to the argument parser.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """

        super(self.__class__, self).add_args(parser)

        parser.add_argument('-o', '--output-config', default=None, help='File path to write an updated MDK config file')
        parser.add_argument('-y', '--no-confirm', default=False, help='No confirmation prompt before file write')

    def action(self, args):
        """
        :param args: The arguments from the command line
        :type args: Namespace
        """

        inputs = InputValues(args, update_keys=False)

        if inputs.obsolete_keys:
            inputs.list_obsolete_keys(fix_warning=False)
            inputs.update_config_keys()
            new_config_fp = inputs.get('output_config') if inputs.get('output_config') else inputs.config_fp

            if inputs.get('no_confirm'):
                inputs.write_config_file(new_config_fp)
            else:
                msg = 'Write updated config file to "{}"?'.format(new_config_fp)
                if inputs.confirm_action(msg):
                    inputs.write_config_file(new_config_fp)
        else:
            self.logger.info('File "{}" is up to date with version {}'.format(
                inputs.config_fp,
                __version__,
            ))


class ConfigCmd(OasisBaseCommand):
    """
    Describes the format of the configuration (JSON) file to use the MDK
    ``model run`` command for running models end-to-end.

    One file will need to be defined per model, usually in the model repository
    and with an indicative name.

    The path-related keys should be strings, given relative to the location of
    configuration file. Optional arguments for a command are usually defaulted
    with appropriate values, but can be overridden by providing a runtime flag.

    :analysis_settings_json: Analysis settings (JSON) file path
    :lookup_data_dir: Model lookup/keys data path (optional)
    :lookup_config_json: Model built-in lookup config. (JSON) file path (optional)
    :lookup_package_dir: Model custom lookup package path (optional)
    :model_version_csv: Model version (CSV) file path (optional)
    :model_data_dir: Model data path
    :model_package_dir: Path to the directory to use as the model specific package (optional)
    :model_run_dir: Model run directory (optional)
    :oed_location_csv: Source OED exposure (CSV) file path
    :oed_accounts_csv: Source OED accounts (CSV) file path (optional)
    :oed_info_csv: Reinsurance (RI) info. file path (optional)
    :oed_scope_csv: RI scope file path (optional)
    :profile_location_csv: Source OED exposure (JSON) profile describing the financial terms contained in the source exposure file (optional)
    :profile_accounts_json: Source OED accouns (JSON) profile describing the financial terms contained in the source accounts file (optional)
    :summarise_exposure: Generates an exposure summary report in JSON
    :ktools_num_processes: The number of concurrent processes used by ktools during model execution - default is ``2``
    :ktools_fifo_relative: Whether to create ktools FIFO queues under the ``./fifo`` subfolder (in the model run directory)
    :ktools_alloc_rule_gul: Override the allocation used in ``fmcalc`` - default is ``1``
    :ktools_alloc_rule_il: Override the allocation used in ``fmcalc`` - default is ``2``
    """
    formatter_class = RawDescriptionHelpFormatter

    sub_commands = {
        'update': ConfigUpdateCmd,
    }



__all__ = [
    'ComputationStep',
]

import os
import pathlib
import logging
import json
import inspect
from collections import OrderedDict

from ..utils.data import get_utctimestamp
from ..utils.exceptions import OasisException


class ComputationStep:
    """
    "Abstract" Class for all Computation Step (ExposurePreAnalysis, GulCalc, ...)
    initialise the object with all specified param un step_param and sub- ComputationStep
    provide a generic interface to get the all those parameter definitions (get_params)

    the Run method must be implemented and contain le business execution logic.

    """

    step_params = []
    chained_commands = []

    def __init__(self, **kwargs):
        """
        initialise the ComputationStep objects:
         - do the basic check for required parameter (required)
         - provide default value if defined (default)
         - check path existence (pre_exist)
         - create necessary directories (is_dir, is_path)
        """
        self.logger = logging.getLogger()
        self.kwargs = kwargs
        self.logger.debug(f"{self.__class__.__name__}: " + json.dumps(self.kwargs, indent=4))

        for param in self.get_params():
            param_value = kwargs.get(param['name'])
            if param_value in [None, ""]:
                if param.get('required'):
                    raise OasisException(f"parameter {param['name']} is required "
                                         f"for Computation Step {self.__class__.__name__}")
                else:
                    param_value = param.get('default')

            if param.get('is_path') and param_value is not None:
                if param.get('pre_exist') and not os.path.exists(param_value):
                    raise OasisException(
                        f"The path {param_value} ({param['help']}) "
                        f"must exist for Computation Step {self.__class__.__name__}")
                else:
                    if param.get('is_dir'):
                        pathlib.Path(param_value).mkdir(parents=True, exist_ok=True)
                    else:
                        pathlib.Path(os.path.dirname(param_value)).mkdir(parents=True, exist_ok=True)
            setattr(self, param['name'], param_value)

    @classmethod
    def get_default_run_dir(cls):
        return os.path.join(os.getcwd(), 'runs', f'{cls.run_dir_key}-{get_utctimestamp(fmt="%Y%m%d%H%M%S")}')

    @classmethod
    def get_params(cls):
        """
        return all the params of the computation step defined in step_params
        and the params from the sub_computation step in chained_commands
        if two params have the same name, return the param definition of the first param found only
        this allow to overwrite the param definition of sub step if necessary.
        """
        params = []
        param_names = set()

        def all_params():
            for param in cls.step_params:
                yield param
            for command in cls.chained_commands:
                for param in command.get_params():
                    yield param

        for param in all_params():
            if param['name'] not in param_names:
                param_names.add(param['name'])
                params.append(param)

        return params

    @classmethod 
    def get_arguments(cls, **kwargs):
        """
        Return a list of default arguments values for the functions parameters 
        If given arg values in 'kwargs' these will override the defaults 
        """
        func_args = {el['name']: el.get('default', None) for el in cls.get_params()}
        for param in kwargs: 
            if param in func_args:
                func_args[param] = kwargs[param]
        return func_args


    @classmethod
    def get_signature(cls):
        """ Create a function signature based on the 'get_params()' return
        """
        try:
            # Create keyword params (without default values)
            params = ["{}=None".format(p.get('name')) for p in cls.get_params() if not p.get('default')]

            # Create keyword params (with default values)
            for p in [p for p in cls.get_params() if p.get('default')]:
                if isinstance(p.get('default'), str):
                    params.append("{}='{}'".format(p.get('name'), p.get('default')))
                elif isinstance(p.get('default'), dict):
                    params.append("{}=dict()".format(p.get('name'), p.get('default')))
                elif isinstance(p.get('default'), OrderedDict):
                    params.append("{}=OrderedDict()".format(p.get('name'), p.get('default')))
                else:
                    params.append("{}={}".format(p.get('name'), p.get('default')))

            exec('def func_sig({}): pass'.format(", ".join(params)))
            return inspect.signature(locals()['func_sig'])
        except Exception:
            # ignore any errors in signature creation and return blank
            return None


    def run(self):
        """method that will be call by all the interface to execute the computation step"""
        raise NotImplemented(f'Methode run must be implemented in {self.__class__.__name__}')



import numba as nb
import numpy as np
import struct
import sys
from numba import jit, int64
from pandas import DataFrame, merge, concat
from typing import Optional, Tuple, Any

from oasislmf.utils.data import merge_dataframes
from .data_access_layers import FileDataAccessLayer
from .descriptors import HeaderTypeDescriptor
from .enums import FileTypeEnum
from .loader_mixin import FileLoader


@jit((int64[:], int64[:], int64[:]), nopython=True)
def define_data_series(vul_ids, inten_ids, damage_bin_maxs):
    vun_list = []

    for i in range(len(vul_ids)):
        vul_id = vul_ids[i]
        inten_id = inten_ids[i]

        for x in range(1, damage_bin_maxs[i] + 1):
            vun_list.append((
                vul_id,
                inten_id,
                x,
                0.0,
            ))
    return vun_list


@jit((int64, int64[:]), nopython=True)
def get_finish_position(identifier, data_vector):
    starting_point = 0
    finishing_point = 0
    match = False

    for i in data_vector:
        if i == identifier:
            match = True
        elif match is True:
            break
        else:
            starting_point += 1
        finishing_point += 1
    return starting_point, finishing_point


class GetModelProcess:
    """
    This class is responsible for loading data from a range of sources and merging them to build a model.

    Attributes:
        data_path (str): the path to the data files needed to construct the model
        model (Optional[DataFrame]): the model for K-tools
        events (Optional[DataFrame]): events preloaded (if None, will load from a file)
        stream_type (int): the type of stream that will be loaded into the header of the stream

    Properties:
        result (Tuple[DataFrame, DataFrame, DataFrame]): the constructed model, empty DataFrame, and damage_bin

    NOTE: This class uses the ModelLoaderMixin so it's data attributes are defined there
    """
    STREAM_HEADER: HeaderTypeDescriptor = HeaderTypeDescriptor()

    def __init__(self, data_path: str, footprint_index_dictionary: dict,
                 events: Optional[DataFrame] = None, file_type: FileTypeEnum = FileTypeEnum.CSV) -> None:
        """
        The constructor for the GetModelProcess class.

        Args:
            data_path: (str) the path to the data files needed to construct the model
            events (Optional[DataFrame]): events preloaded (if None, will load from a file)
            file_type (FileTypeEnum): the file type that the get model loads data from
        """
        self.fdal: FileDataAccessLayer = FileDataAccessLayer(extension=file_type, data_path=data_path)
        self.data_path: str = data_path
        self._vulnerabilities: Optional[FileLoader] = None
        self._footprint: Optional[FileLoader] = None
        self._damage_bin: Optional[FileLoader] = None
        self._events: Optional[FileLoader] = None
        self._items: Optional[FileLoader] = None
        self.model: Optional[DataFrame] = None
        self.events: Optional[Any] = events
        self.stream_type: int = 1
        self.footprint_index_dictionary: dict = footprint_index_dictionary
        self.event_id: int = self.events["event_id"].to_numpy()[0]

    def filter_footprint(self):
        start, stop = self.footprint_index_dictionary[self.event_id]

        self.fdal.footprint.value.insert(loc=0, column='event_id', value=self.event_id)
        filtered_footprint = self.fdal.footprint.value.to_numpy()[start: stop]
        del self.fdal.footprint.value["event_id"]

        if isinstance(self.fdal.items.value, DataFrame):
            self.fdal.items.value = self.fdal.items.value.to_numpy()

        highest_group_id = self.fdal.items.value[-1][4]
        number_of_vulnerability_ids = int(len(self.fdal.items.value) / highest_group_id)
        item_area_peril_ids = sorted([x[2] for x in self.fdal.items.value[0: number_of_vulnerability_ids]])

        item_position_map = {}
        areaperil_id_pointer = 0
        areaperil_id_end_pointer = len(item_area_peril_ids) - 1

        for i in range(0, len(filtered_footprint)):
            if filtered_footprint[i][1] in item_area_peril_ids:
                item_position_map[item_area_peril_ids[areaperil_id_pointer]] = i
                areaperil_id_pointer += 1

            if areaperil_id_pointer > areaperil_id_end_pointer:
                break

        buffer = np.array([[0, 0, 0, 0]])
        for i in sorted(list(item_position_map.keys())):
            row = filtered_footprint[item_position_map[i]: item_position_map[i] + 1]
            buffer = np.concatenate((buffer, row))

        self.model = buffer[1:]

    def merge_vulnerability_with_footprint(self):
        vulnerability_pointer = 0
        end_pointer = len(self.fdal.vulnerabilities.value) - 1

        buffer = np.array([[0, 0, 0, 0, 0, 0, 0, 0]])

        for i in self.model:
            matched = False
            match_ended = False
            footprint_row = i
            position_map = {}
            cached_vulnerability_id: Optional[int] = None

            # generating a dict for matching ascending bin index with vulnerability probabilities
            while match_ended is False:
                vulnerability_row = self.fdal.vulnerabilities.value[vulnerability_pointer]

                if matched is True and footprint_row[2] == vulnerability_row[1]:
                    position_map[vulnerability_row[2]] = vulnerability_row[3]

                elif matched is False and footprint_row[2] == vulnerability_row[1]:
                    cached_vulnerability_id = vulnerability_row[0]
                    position_map[vulnerability_row[2]] = vulnerability_row[3]
                    matched = True

                elif matched is True and footprint_row[2] != vulnerability_row[1]:
                    position_map[vulnerability_row[2]] = vulnerability_row[3]
                    match_ended = True

                if vulnerability_pointer == end_pointer:
                    match_ended = True

                vulnerability_pointer += 1

            cached_bin_index = 1
            cached_probability = 0

            while cached_bin_index <= max(position_map.keys()):
                probability = position_map.get(cached_bin_index, 0.0)
                prob_to = probability * footprint_row[3]
                cached_probability += prob_to

                trimmed_vulnerability_row = np.array(
                    [cached_vulnerability_id, cached_bin_index, probability, cached_probability])
                merged_row = np.concatenate((footprint_row, trimmed_vulnerability_row))
                buffer = np.concatenate((buffer, np.array([merged_row])))
                cached_bin_index += 1

        # event_id	areaperil_id	intensity_bin_id	footprint_probability	vulnerability_id
        # damage_bin_id	vulnerability_probability	prob_to
        self.model = buffer[1:]

    def merge_damage_bin_dict(self):
        damage_map = {}

        for i in self.fdal.damage_bin.value.to_numpy():
            damage_map[i[0]] = i[1:]

        buffer = np.array([[0, 0, 0, 0, 0, 0, 0, 0, 0]])

        for i in self.model:
            interpol = np.array([damage_map[i[5]][2]])
            buffer = np.concatenate((buffer, [np.concatenate((i, interpol))]))

        self.model = buffer[1:]

    def deploy_final_model(self):
        df = DataFrame(self.model, columns=['event_id',
                                            'areaperil_id',
                                            'intensity_bin_id',
                                            'footprint_probability',
                                            'vulnerability_id',
                                            'damage_bin_id',
                                            'vulnerability_probability',
                                            'prob_to',
                                            'interpol'])
        self.model = df

    def print_stream(self) -> None:
        """
        Prints out the stream for cdftocsv.

        Returns: None
        """

        for i in self.model.groupby(["event_id", "areaperil_id", "vulnerability_id"]):

            df = i[1]
            header_row = df.iloc[0]

            sys.stdout.buffer.write(struct.pack("i", int(header_row.event_id)))
            sys.stdout.buffer.write(struct.pack("i", int(header_row.areaperil_id)))
            sys.stdout.buffer.write(struct.pack("i", int(header_row.vulnerability_id)))

            sys.stdout.buffer.write(struct.Struct('i').pack(int(len(df.index))))

            for _, row in df.iterrows():
                sys.stdout.buffer.write(struct.pack("f", float(row.prob_to)))
                sys.stdout.buffer.write(struct.pack("f", float(row.interpol)))  # this is the interpolation

    def run(self) -> None:
        """
        Runs all the functions to construct the model in sequence.

        Returns: None
        """
        if self.should_run is True:
            self.filter_footprint()
            self.merge_vulnerability_with_footprint()
            self.merge_damage_bin_dict()
            self.deploy_final_model()

    @property
    def should_run(self) -> bool:
        if self.footprint_index_dictionary.get(self.event_id) is None:
            return False
        return True

    @property
    def relevant_events(self):
        start, end = self.footprint_index_dictionary[self.event_id]
        return self.fdal.footprint.value.to_numpy()[start: end]



"""
TODO: check it works with csv input (get_mean_damage_bins need to have a csv option)
TODO: work with zipped binary
TODO: have multiple events computed in numba at a time
TODO: use selector and select for output

"""
import argparse
import logging
import numba as nb
import numpy as np
import os
import sys
from contextlib import ExitStack
from logging import NullHandler
from numba.typed import Dict

logger = logging.getLogger(__name__)
logger.addHandler(NullHandler())

buff_size = 65536


areaperil_int = np.dtype(os.environ.get('AREAPERIL_TYPE', 'u4'))
oasis_float = np.dtype(os.environ.get('OASIS_FLOAT', 'f4'))
oasis_int_dtype = np.dtype('i4')
oasis_int = np.int32
oasis_int_size = np.int32().itemsize
buff_int_size = buff_size // oasis_int_size

areaperil_int_relative_size = areaperil_int.itemsize // oasis_int_size
oasis_float_relative_size = oasis_float.itemsize // oasis_int_size
results_relative_size = 2 * oasis_float_relative_size

EventIndexBin = nb.from_dtype(np.dtype([('event_id', np.int32),
                                      ('offset', np.int64),
                                      ('size', np.int64)
                                      ]))

Index_type = nb.from_dtype(np.dtype([('start', np.int64),
                                     ('end', np.int64)
                                      ]))


Event = nb.from_dtype(np.dtype([('areaperil_id', areaperil_int),
                                ('intensity_bin_id', np.int32),
                                ('probability', oasis_float)
                              ]))

event_size = Event.size

footprint_offset = 8

damagebindictionary =  nb.from_dtype(np.dtype([('bin_index', np.int32),
                                    ('bin_from', oasis_float),
                                    ('bin_to', oasis_float),
                                    ('interpolation', oasis_float),
                                    ('interval_type', np.int32),
                                  ]))

damagebindictionaryCsv =  nb.from_dtype(np.dtype([('bin_index', np.int32),
                                                  ('bin_from', oasis_float),
                                                  ('bin_to', oasis_float),
                                                  ('interpolation', oasis_float)]))

EventCSV =  nb.from_dtype(np.dtype([('event_id', np.int32),
                                    ('areaperil_id', areaperil_int),
                                    ('intensity_bin_id', np.int32),
                                    ('probability', oasis_float)
                                  ]))

Item = nb.from_dtype(np.dtype([('id', np.int32),
                               ('coverage_id', np.int32),
                               ('areaperil_id', areaperil_int),
                               ('vulnerability_id', np.int32),
                               ('group_id', np.int32)
                              ]))


Vulnerability = nb.from_dtype(np.dtype([('vulnerability_id', np.int32),
                                        ('intensity_bin_id', np.int32),
                                        ('damage_bin_id', np.int32),
                                        ('probability', oasis_float)
                                       ]))

VulnerabilityIndex = nb.from_dtype(np.dtype([('vulnerability_id', np.int32),
                                             ('offset', np.int64),
                                             ('size', np.int64),
                                             ('original_size', np.int64)
                                            ]))
VulnerabilityRow = nb.from_dtype(np.dtype([('intensity_bin_id', np.int32),
                                           ('damage_bin_id', np.int64),
                                           ('probability', oasis_float)
                                          ]))

vuln_offset = 4

@nb.jit(cache=True)
def load_areaperil_id_u4(int32_mv, cursor, areaperil_id):
    int32_mv[cursor] = areaperil_id.view('i4')
    return cursor + 1


@nb.jit(cache=True)
def load_areaperil_id_u8(int32_mv, cursor, areaperil_id):
    int32_mv[cursor: cursor+1] = areaperil_id.view('i4')
    return cursor + 2


if areaperil_int == 'u4':
    load_areaperil_id = load_areaperil_id_u4
elif areaperil_int == 'u8':
    load_areaperil_id = load_areaperil_id_u8
else:
    raise Exception(f"AREAPERIL_TYPE {areaperil_int} is not implemented chose u4 or u8")


@nb.njit(cache=True)
def get_footprint_idx_from_bin(footprint_idx_bin):
    """
    Gets the footprint index from the footprint binary file.

    Args:
        footprint_idx_bin: (List[EventIndexBin]) data from the footprint.idx file

    Returns: (Tuple[List[int], List[Tuple[int, int]]]) footprint ID to index array. start and finish indexes for footprints
    """
    footprint_idx_dict = np.full(np.max(footprint_idx_bin['event_id']) + 1, np.int32(-1))
    footprint_idx_array = np.empty(footprint_idx_bin.shape[0], dtype=Index_type)

    for i in range(footprint_idx_bin.shape[0]):
        event_idx_bin = footprint_idx_bin[i]
        footprint_idx_dict[event_idx_bin['event_id']] = i
        footprint_idx_array[i]['start'] = (event_idx_bin['offset'] - footprint_offset) // event_size
        footprint_idx_array[i]['end'] = (event_idx_bin['offset'] - footprint_offset + event_idx_bin['size']) // event_size

    return footprint_idx_dict, footprint_idx_array


@nb.jit(cache=True)
def get_footprint_idx_from_csv(footprint_csv):
    """
    Extracts the indexes from the footprint CSV.

    Args:
        footprint_csv: (List[FootPrint]) data loaded from the footprint file.

    Returns: (Tuple[int, bool, List[int], List[Tuple[int, int]], List[FootPrint]]) number of intensity bins,
             if the intensity has uncertainty, footprint index map, start and finish of indexes for the footprints,
             footprint data from the file
    """
    event_count = np.unique(footprint_csv['event_id'])
    num_intensity_bins = max(footprint_csv['intensity_bin_id'])

    footprint_idx_dict = Dict()
    footprint_idx_array = np.empty(len(event_count), dtype=Index_type)
    footprint = np.empty(len(footprint_csv), dtype=Event)
    
    cur_event_id = footprint_csv[0]['event_id']
    cur_areaperil_id = footprint_csv[0]['areaperil_id'] + 1 # make sure cur_areaperil_id is not areaperil_id 
    footprint_idx_dict[cur_event_id] = 0
    footprint_idx_array[0]['start'] = 0

    cur_event_idx = 0

    has_intensity_uncertainty = False

    for i in range(1, len(footprint_csv)):
        event_csv = footprint_csv[i]
        if event_csv['event_id'] != cur_event_id:
            footprint_idx_array[cur_event_idx]['end'] = i - 1 

            cur_event_id = event_csv['event_id']
            cur_event_idx += 1
            footprint_idx_dict[cur_event_id] = cur_event_idx
            footprint_idx_array[cur_event_idx]['start'] = i - 1
        elif cur_areaperil_id == event_csv['areaperil_id']:
            has_intensity_uncertainty == True
        else:
            cur_areaperil_id = event_csv['areaperil_id']

        footprint[i]['areaperil_id'] = event_csv['areaperil_id']
        footprint[i]['intensity_bin_id'] = event_csv['intensity_bin_id']
        footprint[i]['probability'] = event_csv['probability']
    
    footprint_idx_array = footprint_idx_array[1:]
    return num_intensity_bins, has_intensity_uncertainty, footprint_idx_dict, footprint_idx_array, footprint


def get_footprint(static_path, file_type):
    """
    Loads the footprint data from the footprint file and extras meta data around the footprint data.

    Args:
        static_path: (str) the path to the static file housing the footpint data
        file_type: (str) the type of file the footprint file is

    Returns: (Tuple[int, bool, List[int], List[Tuple[int, int]], List[FootPrint]]) number of intensity bins,
             if the intensity has uncertainty, footprint index map, start and finish of indexes for the footprints,
             footprint data from the file
    """
    static_files = set(os.listdir(static_path))

    if "footprint.bin" in static_files and "footprint.idx" in static_files and file_type == "bin":
        with open(os.path.join(static_path, "footprint.bin"), 'rb') as f:
            header = np.frombuffer(f.read(8), 'i4')
            num_intensity_bins = header[0]
            has_intensity_uncertainty = header[1]

        footprint = np.memmap(os.path.join(static_path, "footprint.bin"), dtype=Event, mode='r', offset=footprint_offset)
        footprint_idx_bin = np.memmap(os.path.join(static_path, "footprint.idx"), dtype = EventIndexBin, mode='r')
        footprint_idx_dict, footprint_idx_array = get_footprint_idx_from_bin(footprint_idx_bin)

    elif "footprint.csv" in static_files and file_type == "csv":
        footprint_csv = np.genfromtxt(os.path.join(static_path, "footprint.csv"), dtype=EventCSV, delimiter=",")
        num_intensity_bins, has_intensity_uncertainty, footprint_idx_dict, footprint_idx_array, footprint = get_footprint_idx_from_csv(footprint_csv)
    else:
        raise Exception(f"missing footprint file at {static_path}")

    return num_intensity_bins, has_intensity_uncertainty, footprint_idx_dict, footprint_idx_array, footprint


@nb.jit(cache=True)
def load_items(items):
    """
    Processes the Items loaded from the file extracting meta data around the vulnerability data.

    Args:
        items: (List[Item]) Data loaded from the vulnerability file

    Returns: (Tuple[Dict[int, int], List[int], Dict[int, int], List[Tuple[int, int]], List[int]])
             vulnerability dictionary, vulnerability IDs, areaperil to vulnerability index dictionary,
             areaperil ID to vulnerability index array, areaperil ID to vulnerability array
    """
    areaperil_to_vulns_size = 0
    areaperil_dict = Dict()
    vuln_dict = Dict()
    vuln_idx = 0
    for i in range(items.shape[0]):
        item = items[i]

        # insert the vulnerability index if not in there
        if item['vulnerability_id'] not in vuln_dict:
            vuln_dict[item['vulnerability_id']] = vuln_idx
            vuln_idx += 1

        # insert an area dictionary into areaperil_dict under the key of areaperil ID
        if item['areaperil_id'] not in areaperil_dict:
            area_vuln = Dict()
            area_vuln[item['vulnerability_id']] = 0
            areaperil_dict[item['areaperil_id']] = area_vuln
            areaperil_to_vulns_size += 1
        else:
            if item['vulnerability_id'] not in areaperil_dict[item['areaperil_id']]:
                areaperil_to_vulns_size += 1
                areaperil_dict[item['areaperil_id']][item['vulnerability_id']] = 0

    areaperil_to_vulns_idx_dict = Dict()
    areaperil_to_vulns_idx_array = np.empty(len(areaperil_dict), dtype = Index_type)
    areaperil_to_vulns = np.empty(areaperil_to_vulns_size, dtype = np.int32)
    vulns_id = np.empty(len(vuln_dict), dtype = np.int32)

    for vuln_id, vuln_idx in vuln_dict.items():
        vulns_id[vuln_idx] = vuln_id

    areaperil_i = 0
    vulnerability_i = 0

    for areaperil_id, vulns in areaperil_dict.items():
        areaperil_to_vulns_idx_dict[areaperil_id] = areaperil_i
        areaperil_to_vulns_idx_array[areaperil_i]['start'] = vulnerability_i

        for vuln_id in sorted(vulns):  # sorted is not necessary but doesn't impede the perf and align with cpp getmodel
            areaperil_to_vulns[vulnerability_i] = vuln_dict[vuln_id]
            vulnerability_i +=1
        areaperil_to_vulns_idx_array[areaperil_i]['end'] = vulnerability_i
        areaperil_i+=1

    return vuln_dict, vulns_id, areaperil_to_vulns_idx_dict, areaperil_to_vulns_idx_array, areaperil_to_vulns


def get_items(input_path, file_type):
    """
    Loads the items from the items file.

    Args:
        input_path: (str) the path pointing to the file
        file_type: (str) the type of file being loaded

    Returns: (List[Item]) the Items loaded from the file
    """
    input_files = set(os.listdir(input_path))
    if "items.bin" in input_files and file_type == "bin":
        items = np.memmap(os.path.join(input_path, "items.bin"), dtype=Item, mode='r')
    elif "items.csv" in input_files and file_type == "csv":
        items = np.genfromtxt(os.path.join(input_path, "items.csv"), dtype=Item, delimiter=",")

    return load_items(items)


@nb.njit(cache=True)
def load_vulns_bin_idx(vulns_bin, vulns_idx_bin, vuln_dict,
                                 num_damage_bins, num_intensity_bins):
    """
    Not firing when testing so not able to inspect yet

    Args:
        vulns_bin:
        vulns_idx_bin:
        vuln_dict:
        num_damage_bins:
        num_intensity_bins:

    Returns:
    """
    vuln_array = np.zeros((len(vuln_dict), num_damage_bins, num_intensity_bins), dtype=oasis_float)
    for idx_i in range(vulns_idx_bin.shape[0]):
        vuln_idx = vulns_idx_bin[idx_i]
        if vuln_idx['vulnerability_id'] in vuln_dict:
            cur_vuln_array = vuln_array[vuln_dict[vuln_idx['vulnerability_id']]]
            start = (vuln_idx['offset'] - vuln_offset) // VulnerabilityRow.itemsize
            end = start + vuln_idx['size'] // VulnerabilityRow.itemsize
            for vuln_i in range(start, end):
                vuln = vulns_bin[vuln_i]
                cur_vuln_array[vuln['damage_bin_id'], vuln['intensity_bin_id']] = vuln['probability']

    return vuln_array


@nb.njit(cache=True)
def load_vulns_bin(vulns_bin, vuln_dict, num_damage_bins, num_intensity_bins):
    """
    Loads the vulnerability data grouped by the intensity and damage bins.

    Args:
        vulns_bin: (List[Vulnerability]) vulnerability data from the vulnerability file
        vuln_dict: (Dict[int, int]) maps the vulnerability ID with the index in the vulnerability array
        num_damage_bins: (int) number of damage bins in the data
        num_intensity_bins: (int) the number of intensity bins

    Returns: (List[List[List[floats]]]) vulnerability data grouped by intensity bin and damage bin
    """
    vuln_array = np.zeros((len(vuln_dict), num_damage_bins, num_intensity_bins+1), dtype=oasis_float)
    cur_vulnerability_id = -1

    for vuln_i in range(vulns_bin.shape[0]):
        vuln = vulns_bin[vuln_i]
        if vuln['vulnerability_id'] != cur_vulnerability_id:
            if vuln['vulnerability_id'] in vuln_dict:
                cur_vulnerability_id = vuln['vulnerability_id']
                cur_vuln_array = vuln_array[vuln_dict[cur_vulnerability_id]]
            else:
                cur_vulnerability_id = -1
        if cur_vulnerability_id != -1:
            cur_vuln_array[vuln['damage_bin_id'] - 1, vuln['intensity_bin_id']] = vuln['probability']

    return vuln_array


def get_vulns(static_path, vuln_dict, num_intensity_bins, file_type):
    """
    Loads the vulnerabilities from the file.

    Args:
        static_path: (str) the path pointing to the static file where the data is
        vuln_dict: (Dict[int, int]) maps the vulnerability ID with the index in the vulnerability array
        num_intensity_bins: (int) the number of intensity bins 
        file_type: (str) the type of file being loaded

    Returns: (Tuple[List[List[float]], int]) vulnerability data, number of damage bins
    """
    input_files = set(os.listdir(static_path))
    if "vulnerability.bin" in input_files and file_type == "bin":
        with open(os.path.join(static_path, "vulnerability.bin"), 'rb') as f:
            header = np.frombuffer(f.read(8), 'i4')
            num_damage_bins = header[0]
        if "vulnerability.idx" in static_path:
            vulns_bin = np.memmap(os.path.join(static_path, "vulnerability.bin"), dtype=VulnerabilityRow, offset=4, mode='r')
            vulns_idx_bin = np.memmap(os.path.join(static_path, "vulnerability.idx"), dtype=VulnerabilityIndex, mode='r')
            vuln_array = load_vulns_bin_idx(vulns_bin, vulns_idx_bin, vuln_dict,
                                                      num_damage_bins, num_intensity_bins)
        else:
            vulns_bin = np.memmap(os.path.join(static_path, "vulnerability.bin"), dtype=Vulnerability, offset=4, mode='r')
            vuln_array = load_vulns_bin(vulns_bin, vuln_dict, num_damage_bins, num_intensity_bins)

    elif "vulnerability.csv" in input_files and file_type == "csv":
        vuln_csv = np.genfromtxt(os.path.join(static_path, "vulnerability.csv"), dtype=Vulnerability, delimiter=",")
        num_damage_bins = max(vuln_csv['damage_bin_id'])
        vuln_array = load_vulns_bin(vuln_csv, vuln_dict, num_damage_bins, num_intensity_bins)

    return vuln_array, num_damage_bins


def get_mean_damage_bins(static_path, file_type):
    """
    Loads the mean damage bins from the damage_bin_dict file.

    Args:
        static_path: (str) the path pointing to the static file where the data is
        file_type: (str) the file extension and thus the file type to be loaded

    Returns: (List[Union[damagebindictionaryCsv, damagebindictionary]]) loaded data from the damage_bin_dict file
    """
    file_path = os.path.join(static_path, "damage_bin_dict.bin")
    if not os.path.isfile(str(file_path)) and file_type == "csv":
        return np.genfromtxt(os.path.join(static_path, "damage_bin_dict.csv"), dtype=damagebindictionaryCsv)['interpolation']
    return np.fromfile(os.path.join(static_path, "damage_bin_dict.bin"), dtype=damagebindictionary)['interpolation']


@nb.jit(cache=True, fastmath=True)
def damage_bin_prob(p, intensities_min, intensities_max, vulns, intensities):
    """
    Calculate the probability of an event happening and then causing damage.

    Args:
        p: (float) the probability to be updated
        intensities_min: (int) intensity minimum
        intensities_max: (int) intensity maximum
        vulns: (List[float]) PLEASE FILL IN
        intensities: (List[float]) list of all the intensities

    Returns: (float) the updated probability
    """
    i = intensities_min
    while i < intensities_max:
        p += vulns[i] * intensities[i]
        i += 1
    return p


@nb.jit(cache=True, fastmath=True)
def do_result(vulns_id, vuln_array, mean_damage_bins,
              int32_mv, num_damage_bins,
              intensities_min, intensities_max, intensities,
              event_id, areaperil_id, vuln_i, cursor):
    """
    Calculate the result concerning an event ID.

    Args:
        vulns_id: (List[int]) list of vulnerability IDs
        vuln_array: (List[List[list]]) list of vulnerabilities and their data
        mean_damage_bins: (List[float]) the mean of each damage bin (len(mean_damage_bins) == num_damage_bins)
        int32_mv: (List[int]) FILL IN LATER
        num_damage_bins: (int) number of damage bins in the data
        intensities_min: (int) intensity minimum
        intensities_max: (int) intensity maximum
        intensities: (List[float]) list of all the intensities
        event_id: (int) the event ID that concerns the result being calculated
        areaperil_id: (List[int]) the areaperil ID that concerns the result being calculated
        vuln_i: (int) the index concerning the vulnerability inside the vuln_array
        cursor: (int) PLEASE FILL IN

    Returns: (int) PLEASE FILL IN
    """
    int32_mv[cursor], cursor = event_id, cursor + 1
    int32_mv[cursor:cursor + areaperil_int_relative_size] = areaperil_id.view(oasis_int_dtype)
    cursor += areaperil_int_relative_size
    int32_mv[cursor], cursor = vulns_id[vuln_i], cursor + 1

    cur_vuln_mat = vuln_array[vuln_i]
    p = 0
    cursor_start = cursor
    cursor += 1
    oasis_float_mv = int32_mv[cursor: cursor + num_damage_bins * results_relative_size].view(oasis_float)
    result_cursor = 0
    damage_bin_i = 0

    while damage_bin_i < num_damage_bins:
        p = damage_bin_prob(p, intensities_min, intensities_max, cur_vuln_mat[damage_bin_i], intensities)
        oasis_float_mv[result_cursor], result_cursor = p, result_cursor + 1
        oasis_float_mv[result_cursor], result_cursor = mean_damage_bins[damage_bin_i], result_cursor + 1
        damage_bin_i += 1
        if p > 0.999999940:
            break

    int32_mv[cursor_start] = damage_bin_i
    return cursor + (result_cursor * oasis_float_relative_size)


@nb.njit()
def doCdf(event_id,
          num_intensity_bins, footprint_idx_dict, footprint_idx_array, footprint,
          areaperil_to_vulns_idx_dict, areaperil_to_vulns_idx_array, areaperil_to_vulns,
          vuln_array, vulns_id, num_damage_bins, mean_damage_bins,
          int32_mv, max_result_relative_size):
    """
    Calculates the cumulative distribution function (cdf) for an event ID.

    Args:
        event_id: (int) the event ID the the CDF is being calculated to.
        num_intensity_bins: (int) the number of intensity bins for the CDF
        footprint_idx_dict: (List[int]) maps the event index from the event_id
        footprint_idx_array: (List[Tuple[int, int]]) the index where the event_id starts and the index where the
                                                     event_id finishes
        footprint: (List[Tuple[int, int, float]]) information about the footprint with event_id, areaperil_id,
                                                  probability
        areaperil_to_vulns_idx_dict: (Dict[int, int]) maps the areaperil ID with the ENTER_HERE
        areaperil_to_vulns_idx_array: (List[Tuple[int, int]]) the index where the areaperil ID starts and finishes
        areaperil_to_vulns: (List[int]) maps the areaperil ID to the vulnerability ID
        vuln_array: (List[list]) FILL IN LATER
        vulns_id: (List[int]) list of vulnerability IDs
        num_damage_bins: (int) number of damage bins in the data
        mean_damage_bins: (List[float]) the mean of each damage bin (len(mean_damage_bins) == num_damage_bins)
        int32_mv: (List[int]) FILL IN LATER
        max_result_relative_size: (int) the maximum result size

    Returns: (int)
    """
    if event_id > len(footprint_idx_dict) - 1:
        event_idx_i = -1
    else:
        event_idx_i = footprint_idx_dict[event_id]

    if event_idx_i == -1:
        return 0

    event_idx = footprint_idx_array[event_idx_i]
    intensities_min = num_intensity_bins
    intensities_max = 0
    intensities = np.zeros(num_intensity_bins + 1, dtype=oasis_float)

    areaperil_id = np.zeros(1, dtype=areaperil_int)
    has_vuln = False
    cursor = 0

    for footprint_i in range(event_idx['start'], event_idx['end']):
        event_row = footprint[footprint_i]
        if areaperil_id[0] != event_row['areaperil_id']:
            if has_vuln and intensities_min <= intensities_max:
                areaperil_to_vulns_idx = areaperil_to_vulns_idx_array[areaperil_to_vulns_idx_dict[areaperil_id[0]]]
                intensities_max += 1
                for vuln_idx in range(areaperil_to_vulns_idx['start'], areaperil_to_vulns_idx['end']):
                    vuln_i = areaperil_to_vulns[vuln_idx]
                    if cursor + max_result_relative_size > buff_int_size:
                        yield cursor * oasis_int_size
                        cursor = 0

                    cursor = do_result(vulns_id, vuln_array, mean_damage_bins,
                              int32_mv, num_damage_bins,
                              intensities_min, intensities_max, intensities,
                              event_id, areaperil_id, vuln_i, cursor)

            areaperil_id[0] = event_row['areaperil_id']
            has_vuln = areaperil_id[0] in areaperil_to_vulns_idx_dict

            if has_vuln:
                intensities[intensities_min: intensities_max] = 0
                intensities_min = num_intensity_bins
                intensities_max = 0
        if has_vuln:
            if event_row['probability']>0:
                intensities[event_row['intensity_bin_id']] = event_row['probability']
                if event_row['intensity_bin_id'] > intensities_max:
                    intensities_max = event_row['intensity_bin_id']
                if event_row['intensity_bin_id'] < intensities_min:
                    intensities_min = event_row['intensity_bin_id']

    if has_vuln and intensities_min <= intensities_max:
        areaperil_to_vulns_idx = areaperil_to_vulns_idx_array[areaperil_to_vulns_idx_dict[areaperil_id[0]]]
        intensities_max += 1
        for vuln_idx in range(areaperil_to_vulns_idx['start'], areaperil_to_vulns_idx['end']):
            vuln_i = areaperil_to_vulns[vuln_idx]
            if cursor + max_result_relative_size > buff_int_size:
                yield cursor * oasis_int_size
                cursor = 0

            cursor = do_result(vulns_id, vuln_array, mean_damage_bins,
                      int32_mv, num_damage_bins,
                      intensities_min, intensities_max, intensities,
                      event_id, areaperil_id, vuln_i, cursor)

    yield cursor * oasis_int_size


def run(run_dir, file_in, file_out, file_type):
    """
    Runs the main process of the getmodel process.

    Args:
        run_dir: (str) the directory of where the process is running
        file_in: (Optional[str]) the path to the input directory
        file_out: (Optional[str]) the path to the output directory
        file_type: (str) the type of file being loaded

    Returns: None
    """
    with ExitStack() as stack:
        if file_in is None:
            streams_in = sys.stdin.buffer
        else:
            streams_in = stack.enter_context(open(file_in, 'rb'))

        if file_out is None:
            stream_out = sys.stdout.buffer
        else:
            stream_out = stack.enter_context(open(file_out, 'wb'))

        static_path = os.path.join(run_dir, 'static')
        input_path = os.path.join(run_dir, 'input')

        event_id_mv = memoryview(bytearray(4))
        event_ids = np.ndarray(1, buffer=event_id_mv, dtype='i4')

        vuln_dict, vulns_id, areaperil_to_vulns_idx_dict, areaperil_to_vulns_idx_array, areaperil_to_vulns = get_items(input_path, file_type)

        num_intensity_bins, has_intensity_uncertainty, footprint_idx_dict, footprint_idx_array, footprint = get_footprint(
            static_path, file_type)

        vuln_array, num_damage_bins = get_vulns(static_path, vuln_dict, num_intensity_bins, file_type)

        mean_damage_bins = get_mean_damage_bins(static_path, file_type)

        # even_id, areaperil_id, vulnerability_id, num_result, [oasis_float] * num_result
        max_result_relative_size = 1 + + areaperil_int_relative_size + 1 + 1 + num_damage_bins * results_relative_size

        mv = memoryview(bytearray(buff_size))

        int32_mv = np.ndarray(buff_size // np.int32().itemsize, buffer=mv, dtype=np.int32)

        # header
        stream_out.write(np.uint32(1).tobytes())

        while True:
            len_read = streams_in.readinto(event_id_mv)
            if len_read==0:
                break

            for cursor_bytes in doCdf(event_ids[0],
                  num_intensity_bins, footprint_idx_dict, footprint_idx_array, footprint,
                  areaperil_to_vulns_idx_dict, areaperil_to_vulns_idx_array, areaperil_to_vulns,
                  vuln_array, vulns_id, num_damage_bins, mean_damage_bins,
                                      int32_mv, max_result_relative_size):

                if cursor_bytes:
                    stream_out.write(mv[:cursor_bytes])
                else:
                    break


parser = argparse.ArgumentParser()
parser.add_argument('-i', '--file-in', help='names of the input file_path')
parser.add_argument('-o', '--file-out', help='names of the output file_path')
parser.add_argument('-r', '--run-dir', help='path to the run directory', default='.')
parser.add_argument('-f', '--file-type', help='the type of file to be loaded', default='bin')
parser.add_argument('-v', '--logging-level', help='logging level (debug:10, info:20, warning:30, error:40, critical:50)',
                    default=30, type=int)


def main():
    kwargs = vars(parser.parse_args())

    # add handler to fm logger
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logging_level = kwargs.pop('logging_level')
    logger.setLevel(logging_level)

    run(**kwargs)


if __name__ == "__main__":
    main()






import numba as nb
from typing import Dict, Optional

from .enums import FileTypeEnum
from .file_loader import FileLoader
from .loader_mixin import ModelFileLoaderMixin


class Singleton(type):

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


@nb.jit(cache=True, nopython=True)
def make_footprint_index_dict(footprint_index):
    footprint_map = nb.typed.Dict()
    cached_event_id = footprint_index[0]
    cached_offset = 0

    for i in range(0, len(footprint_index)):
        if footprint_index[i] != cached_event_id:
            footprint_map[cached_event_id] = (cached_offset, i - 1)
            cached_offset = i
            cached_event_id = footprint_index[i]
    return footprint_map


class FileDataAccessLayer(ModelFileLoaderMixin, metaclass=Singleton):

    def __init__(self, data_path, extension: FileTypeEnum = FileTypeEnum.CSV) -> None:
        super().__init__(extension=extension)
        self.data_path: str = data_path
        self._vulnerabilities: Optional[FileLoader] = None
        self._footprint: Optional[FileLoader] = None
        self._damage_bin: Optional[FileLoader] = None
        self._events: Optional[FileLoader] = None
        self._items: Optional[FileLoader] = None
        self.data_path = data_path
        self.extension: FileTypeEnum = extension



import argparse
import numba as nb
import numpy as np
import os
import struct
import sys
from io import StringIO
from pandas import read_csv, DataFrame
from typing import List
from typing import Optional

from .getmodel.enums import FileTypeEnum
from .getmodel.get_model_process import GetModelProcess, FileDataAccessLayer


def _process_input_data() -> Optional[DataFrame]:
    """
    Gets the input from the STDin and converts it to a DataFrame if present.

    Returns: (Optional[DataFrame]) containing event IDs
    """
    data = sys.stdin.buffer.read()

    if data == "":
        return None

    try:
        # data from the evetocsv
        eve_to_csv_data = data.decode()
        return read_csv(StringIO(eve_to_csv_data), sep=",")
    except UnicodeDecodeError:
        pass

    # data directly from eve
    eve_raw_data = [data[i:i + 4] for i in range(0, len(data), 4)]
    eve_buffer = [struct.unpack("i", i)[0] for i in eve_raw_data]

    return DataFrame(eve_buffer, columns=["event_id"])


def _process_file_type(file_type: str) -> FileTypeEnum:
    """
    Extracts the FileTypeEnum type based off of the string.

    Args:
        file_type: (str) the file type to be found

    Returns: (FileTypeEnum) the file type to be found
    """
    enum_map = dict()

    for i in FileTypeEnum:
        enum_map[i.value] = getattr(FileTypeEnum, i.value.upper())

    file_type_value: Optional[FileTypeEnum] = enum_map.get(file_type)
    if file_type_value is None:
        raise ValueError(
            f"file type '{file_type}' is not supported, please pick from {[i.value for i in FileTypeEnum]}"
        )
    return file_type_value


@nb.jit(cache=True, nopython=True)
def make_footprint_index_dict(footprint_index, footprint_offset, event_size):
    """
    Generates the

    :param footprint_index:
    :param footprint_offset:
    :param event_size:
    :return:
    """
    res = nb.typed.Dict()

    for i in range(footprint_index.shape[0]):
        event_index = footprint_index[i]
        res[event_index['event_id']] = ((event_index['offset'] - footprint_offset) // event_size,
                                        (event_index['offset'] - footprint_offset + event_index['size']) // event_size)

    return res


# @nb.jit(cache=True, nopython=True)
def filter_vulnerabilities(items, vulnerabilities):
    highest_group_id = items[-1][4]  # 10
    number_of_vulnerability_ids = int(len(items) / highest_group_id)

    # vulnerability IDs that are in the items
    item_vulnerability_ids = sorted([x[3] for x in items[0: number_of_vulnerability_ids]])
    # item_area_peril_ids = sorted([x[2] for x in items[0: number_of_vulnerability_ids]])

    # generate dictionary that maps the positions of the vulnerabilities
    position_map = nb.typed.Dict()
    matched = False
    vulnerability_id_pointer = 0
    vulnerability_id_end_pointer = len(item_vulnerability_ids) - 1
    start_pointer = 0

    for i in range(0, len(vulnerabilities)):
        if vulnerabilities[i][0] == item_vulnerability_ids[vulnerability_id_pointer] and matched is False:
            matched = True
            start_pointer = i
        elif matched is True and vulnerabilities[i][0] != item_vulnerability_ids[vulnerability_id_pointer]:
            finish_pointer = i
            matched = False
            position_map[item_vulnerability_ids[vulnerability_id_pointer]] = (start_pointer, finish_pointer)
            vulnerability_id_pointer += 1
        if vulnerability_id_pointer > vulnerability_id_end_pointer:
            break

    buffer = np.array([[0.0, 0.0, 0.0, 0.0]])
    for key in sorted(list(position_map.keys())):
        start = position_map[key][0]
        finish = position_map[key][1]
        buffer = np.concatenate((buffer, vulnerabilities[start: finish]))

    return buffer[1:]
    # return position_map


def _generate_footprint_index_dict(file_type: FileTypeEnum, data_path: str) -> dict:
    """
    Reads the footprint binary file creating a footprint index dictionary.

    :return: None
    """
    # fdal is a singleton and will be used in the get_model_process
    fdal = FileDataAccessLayer(extension=file_type, data_path=data_path)
    event_index_struct = nb.from_dtype(np.dtype([('event_id', np.int32),
                                                 ('offset', np.int64),
                                                 ('size', np.int64)
                                                 ]))
    event_struct = nb.from_dtype(np.dtype([('areaperil_id', np.uint32),
                                           ('intensity_bin_id', np.int32),
                                           ('probability', np.float32)
                                           ]))
    footprint_offset = 8
    footprint_buffer: List[str] = fdal.footprint.path.split(".")
    footprint_buffer[-1] = "bin"
    footprint_buffer[-1] = "idx"
    index_path: str = ".".join(footprint_buffer)

    footprint_index = np.memmap(index_path, dtype=event_index_struct, mode="r")

    return make_footprint_index_dict(footprint_index, footprint_offset, event_struct.size)


def _filter_vulnerabilities(file_type: FileTypeEnum, data_path: str):
    fdal = FileDataAccessLayer(extension=file_type, data_path=data_path)
    filtered_vulnerabilities = filter_vulnerabilities(items=fdal.items.value.to_numpy(),
                                                      vulnerabilities=fdal.vulnerabilities.value.to_numpy())
    fdal.vulnerabilities.value = filtered_vulnerabilities


def main() -> None:
    """
    Entry point of the 'getpymodel' command building the module and then piping it out as bytes.

    Returns: None
    """
    # add in argumments that accept the type of file that is being run (CSV, bin, parquet)
    parser = argparse.ArgumentParser(description="Arguments for the get model")
    parser.add_argument("-f", "--file_type", type=str, default="csv")
    args = parser.parse_args()

    data_path: str = str(os.getcwd())
    file_type: FileTypeEnum = _process_file_type(file_type=args.file_type)

    footprint_index_dictionary = _generate_footprint_index_dict(file_type=file_type, data_path=data_path)
    _filter_vulnerabilities(file_type=file_type, data_path=data_path)

    sys.stdout.buffer.write(b'\x01\x00\x00\x00')

    for i in _process_input_data().groupby(["event_id"]):
        process: GetModelProcess = GetModelProcess(data_path=data_path, events=i[1],
                                                   file_type=file_type,
                                                   footprint_index_dictionary=footprint_index_dictionary)

        if process.should_run is True:
            process.run()
            process.print_stream()



__all__ = [
    'VulnerabilityFile',
    'Eventsfile',
    'FootprintBinFile',
    'FootprintIdxFile',
    'DamageBinDictFile',
    'OccurrenceFile',
    'RandomFile',
    'CoveragesFile',
    'ItemsFile',
    'FMProgrammeFile',
    'FMPolicyTCFile',
    'FMProfileFile',
    'FMXrefFile',
    'GULSummaryXrefFile',
    'FMSummaryXrefFile'
]

from collections import OrderedDict
from math import erf
import numpy as np
import os
import struct


class ModelFile:
    """
    Base class for all dummy model files.
    """
    def __init__(self):
        pass

    def seed_rng(self):
        if self.random_seed == 0:
            np.random.seed()
        elif self.random_seed == -1:
            # Add salt to random seed using name of child class
            salt = int.from_bytes(type(self).__name__.encode(), 'little')
            np.random.seed((1234 + salt) % 0xFFFFFFFF)
        else:
            # Add salt to random seed using name of child class
            salt = int.from_bytes(type(self).__name__.encode(), 'little')
            np.random.seed((self.random_seed + salt) % 0xFFFFFFFF)

    def write_file(self):
        """
        Convert data to binary format and writing to file.
        """
        with open(self.file_name, 'wb') as f:
            if self.start_stats:
                for stat in self.start_stats:
                    f.write(struct.pack(stat['dtype'], stat['value']))
            dtypes_list = ''.join(self.dtypes.values())
            for line in self.generate_data():
                f.write(struct.pack('=' + dtypes_list, *(line)))

    def debug_write_file(self):
        """
        Print csv output to screen. Used for debugging file output.
        """
        if self.start_stats:
            for stat in self.start_stats:
                print('{} = {}'.format(stat['desc'], stat['value']))
        line_format = '{}' + ',{}'*(len(self.dtypes)-1)
        print(line_format.format(*self.dtypes.keys()))
        for line in self.generate_data():
            print(line_format.format(*line))


class VulnerabilityFile(ModelFile):
    def __init__(
        self, num_vulnerabilities, num_intensity_bins, num_damage_bins,
        vulnerability_sparseness, random_seed, directory
    ):
        self.num_vulnerabilities = num_vulnerabilities
        self.num_intensity_bins = num_intensity_bins
        self.num_damage_bins = num_damage_bins
        self.vulnerability_sparseness = vulnerability_sparseness
        self.dtypes = OrderedDict([
            ('vulnerability_id', 'i'), ('intensity_bin_index', 'i'),
            ('damage_bin_index', 'i'), ('prob', 'f')
        ])
        self.start_stats = [
            {
                'desc': 'Number of damage bins', 'value': num_damage_bins,
                'dtype': 'i'
            }
        ]
        self.random_seed = random_seed
        self.data_length = num_vulnerabilities * num_intensity_bins * num_damage_bins
        self.file_name = os.path.join(directory, 'vulnerability.bin')

    def generate_data(self):
        super().seed_rng()
        for vulnerability in range(self.num_vulnerabilities):
            for intensity_bin in range(self.num_intensity_bins):

                # Generate probabalities according to vulnerability sparseness
                # and normalise
                triggers = np.random.uniform(size=self.num_damage_bins)
                probabilities = np.apply_along_axis(
                    lambda x: np.where(
                        x < self.vulnerability_sparseness,
                        np.random.uniform(size=x.shape), 0.0
                    ), 0, triggers
                )
                probabilities /= np.sum(probabilities)

                for damage_bin, probability in enumerate(probabilities):
                    yield vulnerability+1, intensity_bin+1, damage_bin+1, probability


class EventsFile(ModelFile):
    def __init__(self, num_events, directory):
        self.num_events = num_events
        self.dtypes = {'event_id': 'i'}
        self.start_stats = None
        self.data_length = num_events
        self.file_name = os.path.join(directory, 'events.bin')

    def generate_data(self):
        return (tuple([event]) for event in range(1, self.num_events+1))


class FootprintFiles(ModelFile):
    bin_dtypes = OrderedDict([
        ('areaperil_id', 'i'), ('intensity_bin_id', 'i'), ('probability', 'f')
    ])

    def __init__(
        self, num_events, num_areaperils, areaperils_per_event,
        num_intensity_bins, intensity_sparseness, no_intensity_uncertainty
    ):
        self.num_events = num_events
        self.num_areaperils = num_areaperils
        self.areaperils_per_event = areaperils_per_event
        self.num_intensity_bins = num_intensity_bins
        self.intensity_sparseness = intensity_sparseness
        self.no_intensity_uncertainty = no_intensity_uncertainty

    def get_bin_start_stats(self):
        return [
            {
                'desc': 'Number of intensity bins',
                'value': self.num_intensity_bins, 'dtype': 'i'
            },
            {
                'desc': 'Has Intensity Uncertainty',
                'value': not self.no_intensity_uncertainty, 'dtype': 'i'
            }
        ]


class FootprintBinFile(FootprintFiles):
    def __init__(
        self, num_events, num_areaperils, areaperils_per_event,
        num_intensity_bins, intensity_sparseness, no_intensity_uncertainty,
        random_seed, directory
    ):
        super().__init__(
            num_events, num_areaperils, areaperils_per_event,
            num_intensity_bins, intensity_sparseness, no_intensity_uncertainty
        )
        self.start_stats = self.get_bin_start_stats()
        self.dtypes = FootprintFiles.bin_dtypes
        self.random_seed = random_seed
        if no_intensity_uncertainty:
            self.data_length = num_events * areaperils_per_event
        else:
            self.data_length = num_events * areaperils_per_event * num_intensity_bins
        self.file_name = os.path.join(directory, 'footprint.bin')

    def generate_data(self):
        super().seed_rng()
        for event in range(self.num_events):

            if self.areaperils_per_event == self.num_areaperils:
                selected_areaperils = np.arange(1, self.num_areaperils+1)
            else:
                selected_areaperils = np.random.choice(
                    self.num_areaperils, self.areaperils_per_event,
                    replace=False
                )
                selected_areaperils += 1
                selected_areaperils = np.sort(selected_areaperils)

            for areaperil in selected_areaperils:
                if self.no_intensity_uncertainty:
                    intensity_bin = np.random.randint(
                        1, self.num_intensity_bins+1
                    )
                    probability = 1.0
                    yield areaperil, intensity_bin, probability
                else:
                    # Generate probabalities according to intensity sparseness
                    # and normalise
                    triggers = np.random.uniform(size=self.num_intensity_bins)
                    probabilities = np.apply_along_axis(
                        lambda x: np.where(
                            x < self.intensity_sparseness,
                            np.random.uniform(size=x.shape), 0.0
                        ), 0, triggers
                    )
                    probabilities /= np.sum(probabilities)

                    for intensity_bin, probability in enumerate(probabilities):
                        yield areaperil, intensity_bin+1, probability


class FootprintIdxFile(FootprintFiles):
    def __init__(
        self, num_events, num_areaperils, areaperils_per_event,
        num_intensity_bins, intensity_sparseness, no_intensity_uncertainty,
        directory
    ):
        super().__init__(
            num_events, num_areaperils, areaperils_per_event,
            num_intensity_bins, intensity_sparseness, no_intensity_uncertainty
        )
        self.start_stats = None
        self.dtypes = OrderedDict([
            ('event_id', 'i'), ('offset', 'q'), ('size', 'q')
        ])
        self.data_length = num_events
        self.file_name = os.path.join(directory, 'footprint.idx')

    def generate_data(self):
        # Size is the same for all events
        size = 0
        for dtype in FootprintFiles.bin_dtypes.values():
            size += struct.calcsize(dtype)
        size *= self.areaperils_per_event
        if not self.no_intensity_uncertainty:
            size *= self.num_intensity_bins
        # Set initial offset
        offset = 0
        for stat in self.get_bin_start_stats():
            offset += struct.calcsize(stat['dtype'])

        for event in range(self.num_events):
            yield event+1, offset, size
            offset += size


class DamageBinDictFile(ModelFile):
    def __init__(self, num_damage_bins, directory):
        self.num_damage_bins = num_damage_bins
        self.dtypes = OrderedDict([
            ('bin_index', 'i'), ('bin_from', 'f'), ('bin_to', 'f'),
            ('interpolation', 'f'), ('interval_type', 'i')
        ])
        self.start_stats = None
        self.data_length = num_damage_bins
        self.file_name = os.path.join(directory, 'damage_bin_dict.bin')

    def generate_data(self):
        # Exclude first and last bins for now
        bin_indexes = np.arange(self.num_damage_bins-2)
        bin_from_values = bin_indexes / (self.num_damage_bins-2)
        bin_to_values = (bin_indexes + 1) / (self.num_damage_bins-2)
        # Set interpolation in middle of bin
        interpolations = (0.5 + bin_indexes) / (self.num_damage_bins-2)
        # Insert first and last bins
        bin_indexes += 2
        bin_indexes = np.insert(bin_indexes, 0, 1)
        bin_indexes = np.append(bin_indexes, self.num_damage_bins)
        fields = [bin_from_values, bin_to_values, interpolations]
        for i, field in enumerate(fields):
            fields[i] = np.insert(field, 0, 0)
            fields[i] = np.append(fields[i], 1)
        bin_from_values, bin_to_values, interpolations = fields
        # Set interval type for all bins to 0 (unused)
        interval_type = 0

        for bin_id, bin_from, bin_to, interpolation in zip(
            bin_indexes, bin_from_values, bin_to_values, interpolations
        ):
            yield bin_id, bin_from, bin_to, interpolation, interval_type


class OccurrenceFile(ModelFile):
    def __init__(
            self, num_events, num_periods, random_seed, directory, mean, stddev
    ):
        self.num_events = num_events
        self.num_periods = num_periods
        self.dtypes = OrderedDict([
            ('event_id', 'i'), ('period_no', 'i'), ('occ_date_id', 'i')
        ])
        self.date_algorithm = 1
        self.start_stats = [
            {
                'desc': 'Date algorithm', 'value': self.date_algorithm,
                'dtype': 'i'
            },
            {
                'desc': 'Number of periods', 'value': self.num_periods,
                'dtype': 'i'
            }
        ]
        self.random_seed = random_seed
        self.data_length = num_events
        self.mean = mean
        self.stddev = stddev
        self.file_name = os.path.join(directory, 'occurrence.bin')

    def get_num_periods_from_truncated_normal_cdf(self):
        # Truncated Normal Cumulative Distribution Function:
        # F(x) = [Phi(g(x)) - Phi(g(a))] / [Phi(g(b)) - Phi(g(a))]
        # g(y) = (y - mean) / standard_deviation
        # Phi(g(y)) = 1/2 * (1 + erf(g(y) / sqrt(2)))
        # a = lower boundary = 0.5, b = upper boundary = infinity
        #   therefore g(b) -> infinity ===> Phi(g(b)) -> 1
        alpha = (0.5 - self.mean) / self.stddev
        phi_alpha = 0.5 * (1 + erf(alpha / np.sqrt(2)))
        rand_no = np.random.random()
        bound_a = 0.5
        while True:
            xi = (bound_a - self.mean) / self.stddev
            phi_xi = 0.5 * (1 + erf(xi / np.sqrt(2)))
            if rand_no < ((phi_xi - phi_alpha) / (1 - phi_alpha)):
                return int(bound_a)
            bound_a += 1

    def get_num_periods_per_event(self):
        # Return mean if standard deviation is 0
        if self.stddev == 0:
            return self.mean
        else:
            return self.get_num_periods_from_truncated_normal_cdf()

    def set_occ_date_id(self, year, month, day):
        # Set date relative to epoch
        month = (month + 9) % 12
        year = year - month // 10
        return 365 * year + year // 4 - year // 100 + year // 400 + (306 * month + 5) // 10 + (day - 1)

    def generate_data(self):
        super().seed_rng()
        months = np.arange(1, 13)
        days_per_month = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
        months_weights = np.array(days_per_month, dtype=float)
        months_weights /= months_weights.sum()   # Normalise
        for event in range(self.num_events):
            for _ in range(self.get_num_periods_per_event()):
                period_no = np.random.randint(1, self.num_periods+1)
                occ_year = period_no   # Assume one period represents one year
                occ_month = np.random.choice(months, p=months_weights)
                occ_day = np.random.randint(1, days_per_month[occ_month-1])
                occ_date = self.set_occ_date_id(occ_year, occ_month, occ_day)
                yield event+1, period_no, occ_date


class RandomFile(ModelFile):
    def __init__(self, num_randoms, random_seed, directory):
        self.num_randoms = num_randoms
        self.dtypes = {'random_no': 'f'}
        self.start_stats = None
        self.random_seed = random_seed
        self.data_length = num_randoms
        self.file_name = os.path.join(directory, 'random.bin')

    def generate_data(self):
        super().seed_rng()
        # First random number is 0
        return (tuple([np.random.uniform()]) if i != 0 else (0,) for i in range(self.num_randoms))


class CoveragesFile(ModelFile):
    def __init__(
        self, num_locations, coverages_per_location, random_seed, directory
    ):
        self.num_locations = num_locations
        self.coverages_per_location = coverages_per_location
        self.dtypes = {'tiv': 'f'}
        self.start_stats = None
        self.random_seed = random_seed
        self.data_length = num_locations * coverages_per_location
        self.file_name = os.path.join(directory, 'coverages.bin')

    def generate_data(self):
        super().seed_rng()
        # Assume 1-1 mapping between item and coverage IDs
        return (
            tuple([np.random.uniform(1, 1000000)]) for _ in range(
                self.num_locations * self.coverages_per_location
            )
        )


class ItemsFile(ModelFile):
    def __init__(
        self, num_locations, coverages_per_location, num_areaperils,
        num_vulnerabilities, random_seed, directory
    ):
        self.num_locations = num_locations
        self.coverages_per_location = coverages_per_location
        self.num_areaperils = num_areaperils
        self.num_vulnerabilities = num_vulnerabilities
        self.dtypes = OrderedDict([
            ('item_id', 'i'), ('coverage_id', 'i'), ('areaperil_id', 'i'),
            ('vulnerability_id', 'i'), ('group_id', 'i')
        ])
        self.start_stats = None
        self.random_seed = random_seed
        self.data_length = num_locations * coverages_per_location
        self.file_name = os.path.join(directory, 'items.bin')

    def generate_data(self):
        super().seed_rng()
        for location in range(self.num_locations):
            areaperils = np.random.randint(
                1, self.num_areaperils+1, size=self.coverages_per_location
            )
            vulnerabilities = np.random.randint(
                1, self.num_vulnerabilities+1, size=self.coverages_per_location
            )
            for coverage in range(self.coverages_per_location):
                item = self.coverages_per_location * location + coverage + 1
                # Assume 1-1 mapping between item and coverage IDs
                # Assume group ID mapped to location
                yield item, item, areaperils[coverage], vulnerabilities[coverage], location+1


class FMFile(ModelFile):
    def __init__(self, num_locations, coverages_per_location):
        self.num_locations = num_locations
        self.coverages_per_location = coverages_per_location
        self.start_stats = None


class FMProgrammeFile(FMFile):
    def __init__(self, num_locations, coverages_per_location, directory):
        super().__init__(num_locations, coverages_per_location)
        self.dtypes = OrderedDict([
            ('from_agg_id', 'i'), ('level_id', 'i'), ('to_agg_id', 'i')
        ])
        self.data_length = num_locations * coverages_per_location * 2   # 2 from number of levels
        self.file_name = os.path.join(directory, 'fm_programme.bin')

    def generate_data(self):
        levels = [1, 10]
        levels = range(1, len(levels)+1)
        for level in levels:
            for agg_id in range(
                1, self.num_locations * self.coverages_per_location + 1
            ):
                # Site coverage FM level
                if level == 1:
                    yield agg_id, level, agg_id
                # Policy layer FM level
                elif level == len(levels):
                    yield agg_id, level, 1


class FMPolicyTCFile(FMFile):
    def __init__(
        self, num_locations, coverages_per_location, num_layers, directory
    ):
        super().__init__(num_locations, coverages_per_location)
        self.num_layers = num_layers
        self.dtypes = OrderedDict([
            ('layer_id', 'i'), ('level_id', 'i'), ('agg_id', 'i'),
            ('policytc_id', 'i')
        ])
        self.data_length = num_locations * coverages_per_location + num_layers
        self.file_name = os.path.join(directory, 'fm_policytc.bin')

    def generate_data(self):
        # Site coverage #1 & policy layer #10 FM levels
        levels = [1, 10]
        levels = range(1, len(levels)+1)
        policytc_id = 1
        for level in levels:
            # Site coverage FM level
            if level == 1:
                for agg_id in range(
                    1, self.num_locations * self.coverages_per_location + 1
                ):
                    # One layer in site coverage FM level
                    yield level, agg_id, 1, policytc_id
                policytc_id += 1   # Next policytc_id
            # Policy layer FM level
            elif level == len(levels):
                for layer in range(self.num_layers):
                    yield level, 1, layer+1, policytc_id
                    policytc_id += 1   # Next policytc_id


class FMProfileFile(ModelFile):
    def __init__(self, num_layers, directory):
        self.num_layers = num_layers
        self.dtypes = OrderedDict([
            ('policytc_id', 'i'), ('calcrule_id', 'i'), ('deductible1', 'f'),
            ('deductible2', 'f'), ('deductible3', 'f'), ('attachment1', 'f'),
            ('limit1', 'f'), ('share1', 'f'), ('share2', 'f'), ('share3', 'f')
        ])
        self.start_stats = None
        self.data_length = 1 + num_layers   # 1 from pass through at level 1
        self.file_name = os.path.join(directory, 'fm_profile.bin')

    def generate_data(self):
        # Pass through for level 1
        profile_rows = [(1, 100, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0)]
        # First policy
        init_policytc_id = 2
        init_attachment1 = 500000.0
        attachment1_offset = 5000000.0
        max_limit1 = 100000000.0
        for layer in range(self.num_layers):
            policytc_id = init_policytc_id + layer
            attachment1 = init_attachment1 + attachment1_offset * layer
            # Set limit1 at maximum for last layer
            if (layer+1) == self.num_layers:
                limit1 = max_limit1
            else:
                limit1 = attachment1_offset * (layer+1)
            profile_rows.append(
                (policytc_id, 2, 0.0, 0.0, 0.0, attachment1, limit1, 0.3, 0.0, 0.0)
            )
        for row in profile_rows:
            yield row


class FMXrefFile(FMFile):
    def __init__(
        self, num_locations, coverages_per_location, num_layers, directory
    ):
        super().__init__(num_locations, coverages_per_location)
        self.num_layers = num_layers
        self.dtypes = OrderedDict([
            ('output', 'i'), ('agg_id', 'i'), ('layer_id', 'i')
        ])
        self.data_length = num_locations * coverages_per_location * num_layers
        self.file_name = os.path.join(directory, 'fm_xref.bin')

    def generate_data(self):
        layers = range(1, self.num_layers+1)
        output_count = 1
        for agg_id in range(
            1, self.num_locations * self.coverages_per_location + 1
        ):
            for layer in layers:
                yield output_count, agg_id, layer
                output_count += 1


class GULSummaryXrefFile(FMFile):
    def __init__(self, num_locations, coverages_per_location, directory):
        super().__init__(num_locations, coverages_per_location)
        self.dtypes = OrderedDict([
            ('item_id', 'i'), ('summary_id', 'i'), ('summaryset_id', 'i')
        ])
        self.data_length = num_locations * coverages_per_location
        self.file_name = os.path.join(directory, 'gulsummaryxref.bin')

    def generate_data(self):
        summary_id = 1
        summaryset_id = 1
        for item in range(self.num_locations * self.coverages_per_location):
            yield item+1, summary_id, summaryset_id


class FMSummaryXrefFile(FMFile):
    def __init__(
        self, num_locations, coverages_per_location, num_layers, directory
    ):
        super().__init__(num_locations, coverages_per_location)
        self.num_layers = num_layers
        self.dtypes = OrderedDict([
            ('output_id', 'i'), ('summary_id', 'i'), ('summaryset_id', 'i')
        ])
        self.data_length = num_locations * coverages_per_location * num_layers
        self.file_name = os.path.join(directory, 'fmsummaryxref.bin')

    def generate_data(self):
        summary_id = 1
        summaryset_id = 1
        for output_id in range(
            self.num_locations * self.coverages_per_location * self.num_layers
        ):
            yield output_id+1, summary_id, summaryset_id



__all__ = [
    'oasis_log',
    'read_log_config',
    'set_rotating_logger'
]

"""
Logging utils.
"""
import inspect
import logging
import os
import time

from functools import wraps
from logging.handlers import RotatingFileHandler


def getargspec(func):
    if hasattr(inspect, 'getfullargspec'):
        return inspect.getfullargspec(func)
    else:
        return inspect.getargspec(func)


def set_rotating_logger(
    log_file_path=inspect.stack()[1][1],
    log_level=logging.INFO,
    max_file_size=10**7,
    max_backups=5
):
    _log_fp = log_file_path
    if not os.path.isabs(_log_fp):
        _log_fp = os.path.abspath(_log_fp)

    log_dir = os.path.dirname(_log_fp)
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)

    handler = RotatingFileHandler(
        _log_fp,
        maxBytes=max_file_size,
        backupCount=max_backups
    )

    logging.getLogger().setLevel(log_level)
    logging.getLogger().addHandler(handler)
    formatter = logging.Formatter(
        "%(asctime)s - %(levelname)s - %(message)s")

    handler.setFormatter(formatter)


def read_log_config(config_parser):
    """
    Read an Oasis standard logging config
    """
    log_file = config_parser['LOG_FILE']
    log_level = config_parser['LOG_LEVEL']
    log_max_size_in_bytes = int(config_parser['LOG_MAX_SIZE_IN_BYTES'])
    log_backup_count = int(config_parser['LOG_BACKUP_COUNT'])

    log_dir = os.path.dirname(log_file)
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)

    handler = RotatingFileHandler(
        log_file, maxBytes=log_max_size_in_bytes,
        backupCount=log_backup_count)
    logging.getLogger().setLevel(log_level)
    logging.getLogger().addHandler(handler)
    formatter = logging.Formatter(
        "%(asctime)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)


def oasis_log(*args, **kwargs):
    """
    Decorator that logs the entry, exit and execution time.
    """
    logger = logging.getLogger()

    def actual_oasis_log(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            func_name = func.__name__
            caller_module_name = func.__globals__.get('__name__')

            if func_name == '__init__':
                logger.debug("RUNNING: {}.{}".format(
                    caller_module_name, func_name))
            else:
                logger.info("RUNNING: {}.{}".format(
                    caller_module_name, func_name))

            args_name = getargspec(func)[0]
            args_dict = dict(zip(args_name, args))

            for key, value in args_dict.items():
                if key == "self":
                    continue
                logger.debug("    {} == {}".format(key, value))

            if len(args) > len(args_name):
                for i in range(len(args_name), len(args)):
                    logger.debug("    {}".format(args[i]))
            for key, value in kwargs.items():
                logger.debug("    {} == {}".format(key, value))

            start = time.time()
            result = func(*args, **kwargs)
            end = time.time()

            # Only log timestamps on functions which took longer than 10ms
            if (end-start) > 0.01:
                logger.info(
                    "COMPLETED: {}.{} in {}s".format(
                        caller_module_name, func_name, round(end - start, 2)))
            else:
                logger.debug(
                    "COMPLETED: {}.{} in {}s".format(
                        caller_module_name, func_name, round(end - start, 2)))
            return result
        return wrapper

    if len(args) == 1 and callable(args[0]):
        return actual_oasis_log(args[0])
    else:
        return actual_oasis_log



from .descriptors import FileMapDescriptor
from .enums import FileTypeEnum
from .file_loader import FileLoader


class ModelFileLoaderMixin:
    """
    This Mixin class is responsible for loading data for the get model.
    """
    FILE_MAP = FileMapDescriptor()

    def __init__(self, extension: FileTypeEnum = FileTypeEnum.CSV) -> None:
        """
        The constructor for the ModelLoaderMixin class.

        Args:
            extension: (FileTypeEnum) type of extension to be used
        """
        self.extension: FileTypeEnum = extension

    def load_data_if_none(self, name: str) -> None:
        """
        Loads the data from the file in the directory of the self.data_path setting the data to the self._{name}
        attribute.

        Args:
            name: (str) this is doing to be used to get the name of the file using the self.FILE_MAP

        Returns: None
        """
        if getattr(self, f"_{name}") is None:
            file_handler: FileLoader = FileLoader(file_path=self.data_path + f"/{self.FILE_MAP[name]}",
                                                  label=name)
            setattr(self, f"_{name}", file_handler)

    @property
    def items(self) -> FileLoader:
        self.load_data_if_none(name="items")
        return self._items

    @property
    def vulnerabilities(self) -> FileLoader:
        self.load_data_if_none(name="vulnerabilities")
        return self._vulnerabilities

    @property
    def footprint(self) -> FileLoader:
        self.load_data_if_none(name="footprint")
        return self._footprint

    @property
    def damage_bin(self) -> FileLoader:
        self.load_data_if_none(name="damage_bin")
        return self._damage_bin

    @property
    def events(self) -> FileLoader:
        self.load_data_if_none(name="events")
        return self._events

    @items.setter
    def items(self, value) -> None:
        self._items = value

    @vulnerabilities.setter
    def vulnerabilities(self, value) -> None:
        self._vulnerabilities = value

    @footprint.setter
    def footprint(self, value) -> None:
        self._footprint = value

    @damage_bin.setter
    def damage_bin(self, value) -> None:
        self._damage_bin = value

    @events.setter
    def events(self, value) -> None:
        if value is not None:
            placeholder = FileLoader(file_path=self.data_path + f"/{self.FILE_MAP['events']}",
                                     label="events")
            placeholder.value = value
            value = placeholder
        self._events = value
